
/*
   This file is part of AirFart.

   AirFart is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   AirFart is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with AirFart; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "framesniffer.h"
#include "printchars.h"

/****************************************************************************/
void frame_sniffer::start_sniffing()
{
	if( -1 == go_monitor_mode() )
		return;	
	if( -1 == set_socket() )
		return;
	if( -1 == bind_socket() )
		return;

	unsigned int bytes_read, i, j;
	unsigned int ssi_type, avs_header_len;
	int ssi_signal, ssi_noise;
	char src_ssid[255] = "not_implemented_yet";
	unsigned char src_eth_addr[ETH_ADDR_LEN];
	unsigned char dst_eth_addr[ETH_ADDR_LEN];
	unsigned char bcast_eth_addr[ETH_ADDR_LEN];
	unsigned char dot11_ident[DOT11_IDENT_LEN] = DOT11_IDENT;
	memset( bcast_eth_addr, 0xff, ETH_ADDR_LEN );
	int src_eth_addr_offset;

	#ifdef DEBUG_SNIFFER
		printf( "Reading from device...\n" );
	#endif

	bytes_read = 0;
	while( bytes_read != -1 )
	{
		bytes_read = recv (skt, msgbuf, MAX_BUFFER_SIZE, 0 );

		if( bytes_read == -1 )
		{
			printf( "Could not read from device. Sleeping for retry!\n" );
			sleep(5);
			continue;
		}
		
		#ifdef DEBUG_SNIFFER
			printf( "__________________________________________________________________\n" );
			printf( "Captured a frame: %d bytes\n", bytes_read );
			print_chars( "Frame contents:",  msgbuf, bytes_read );
		#endif

		/* error checking */

		/* Check for the 802.11 ident value at the beginning of the packet */
		if( 0 != memcmp( msgbuf+DOT11_IDENT_OFFSET, dot11_ident, DOT11_IDENT_LEN ) )
		{
			#ifdef DEBUG_SNIFFER
				printf( " -- Error: this is not an 802.11 packet --\n" );
			#endif
			continue;
		}
			
		/* Extract the AVS header len from the wlan header */
		memcpy( &avs_header_len, msgbuf+AVS_HEADER_LEN_OFFSET, 4 );
		avs_header_len = ntohl( avs_header_len );

		/* If the extracted avs header len isn't 64, we've got a malformed
		 * packet on our hands (or the card isn't passing the wlan header */
		if( avs_header_len != AVS_HEADER_LEN )
		{
			#ifdef DEBUG_SNIFFER
				printf( " -- Error in AVS header length -- \n" );
				printf( " -- Expected %d, got %d -- \n", AVS_HEADER_LEN, avs_header_len );
			#endif
				
			continue;
		}

		/* We are expecting the ethernet frame to be big enough to
		 * have the source eth addr in it. If it ain't, it's an error */
		if( bytes_read < SRC_ETH_ADDR_OFFSET+ETH_ADDR_LEN )
		{
			#ifdef DEBUG_SNIFFER
				printf( " -- Error: frame appears malformed (too short) -- \n" );
			#endif
				
			continue;
		}

		/* We are expecting the ethernet frame to be big enough to
		 * have the destination eth addr in it. If it ain't, it's an error */
		if( bytes_read < DST_ETH_ADDR_OFFSET+ETH_ADDR_LEN )
		{	
			#ifdef DEBUG_SNIFFER
				printf( " -- Error: frame appears malformed (too short) -- \n" );
			#endif
				
			continue;
		}

		/* Extract the signal data */
		//memcpy( &ssi_type, msgbuf+SSI_TYPE_OFFSET, 4 ); // see doc for type descs
		memcpy( &ssi_signal, msgbuf+SSI_SIGNAL_OFFSET, 4 ); // strength
		//memcpy( &ssi_noise, msgbuf+SSI_NOISE_OFFSET, 4 ); // noise (almost always 0)

		/* is this a special empty 802.11 frame? */
		/*
		if( bytes_read == EMPTY_FRAME_LEN )
		{
			src_eth_addr_offset  = EMPTY_FRAME_SRC_ETH_ADDR_OFFSET;
		}
		else
		{
			src_eth_addr_offset = SRC_ETH_ADDR_OFFSET;
		}
		*/
		
		/* Source and Dest ethernet addresses */
		memcpy( dst_eth_addr, msgbuf+DST_ETH_ADDR_OFFSET, ETH_ADDR_LEN );
		memcpy( src_eth_addr, msgbuf+SRC_ETH_ADDR_OFFSET, ETH_ADDR_LEN );

		/* This data is in network format, convert it to the local machine's
		 * endian'ness */
		//ssi_type = ntohl( ssi_type );
		ssi_signal= ntohl( ssi_signal );
		//ssi_noise = ntohl( ssi_noise );
		
		#ifdef DEBUG_SNIFFER
			printf( " src eth: " );
			print_eth_addr( src_eth_addr );
			printf( "\n" );
			printf( " dst eth: " );
			print_eth_addr( dst_eth_addr );
			printf( "\n" );
			printf( " strength: %d\n", ssi_signal );
		#endif
			
		/* Pass the eth address and sig strength to our stats collector */
		stats->add_packet( src_eth_addr, src_ssid, ssi_signal );
		
	}

}

/********************************************************************************/
int frame_sniffer::go_monitor_mode()
{
	char monitor_command[255];
	int wlan_ret_val;
	
	#ifdef DEBUG_SNIFFER
		printf( "Putting device in monitor mode... " );
	#endif

	// just use DEFAULT_CHANNEL for now -- learn how to scan multiple channels later
	sprintf( monitor_command, "wlanctl-ng wlan0 lnxreq_wlansniff enable=true channel=%d wlanheader=true;", DEFAULT_CHANNEL );
	sprintf( monitor_command, "%s ifconfig wlan0 up", monitor_command );

	#ifndef DEBUG_SNIFFER
		sprintf( monitor_command, "%s %s", monitor_command, " > /dev/null" );
	#endif
	
	wlan_ret_val = system( monitor_command );
	if( wlan_ret_val != 0 )
	{
		printf( "Could not put the card in monitor mode!\n" );
		printf( "  Either the device is not present or 'wlanctl-ng' is not in your path\n" );
		return -1;
	}
	else
	{
		#ifdef DEBUG_SNIFFER
			printf( "Done.\n" );
		#endif
		return 0;
	}
	
}

/********************************************************************************/
int frame_sniffer::set_socket()
{
	#ifdef DEBUG_SNIFFER
		printf( "Opening raw socket to device... " );
	#endif
		
	skt = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (skt == -1)
	{
		printf("ERROR: Could not open connection to the network card!\n");
		return -1;
	}
	else
	{
		#ifdef DEBUG_SNIFFER
			printf( "Done.\n" );
		#endif
		return 0;
	}
}

/********************************************************************************/
int frame_sniffer::bind_socket()
{
	#ifdef DEBUG_SNIFFER
		printf( "Binding socket to device... " );
	#endif
	int bind_result;
	struct sockaddr_ll sll;

	memset(&sll, 0, sizeof(sll));
	sll.sll_family          = AF_PACKET;
	sll.sll_ifindex         = 0;
	sll.sll_protocol        = htons(ETH_P_ALL);

	bind_result = bind( skt, (struct sockaddr *) &sll, sizeof(sll) );
	if( bind_result == -1 )
	{
		printf( "ERROR: Could not bind the socket!\n" );
		return -1;
	}
	else
	{
		#ifdef DEBUG_SNIFFER
			printf( "Done.\n" );
		#endif

		return 0;
	}
}

/********************************************************************************/
void frame_sniffer::print_eth_addr( unsigned char *eth_addr )
{
	printf( "%x:%x:%x:%x:%x:%x",
		eth_addr[0],eth_addr[1],eth_addr[2],
		eth_addr[3],eth_addr[4],eth_addr[5] );
}

/********************************************************************************/
void frame_sniffer::set_stats_collector( Stats *s)
{
	stats = s;
}

